# -*- coding: utf-8 -*-
try:
    from tqdm import tqdm
    tqdm_ = tqdm(desc='↑', unit='B', unit_scale=True)  
except:
    tqdm_ = None
def report(current,max_val):
    if tqdm_ is None:return
    tqdm_.total = max_val
    tqdm_.n = current    
    tqdm_.update(0)
    tqdm_.refresh()
def close():
    if tqdm_ is None:return
    tqdm_.close()


# -*- coding: utf-8 -*-
import os
from sys import path
from . import DownloadResult
__desc__ = '本地文件'
__cfg_help__ = '''
    cover (str) - 封面图片路径
e.g. --localfile "le videos/" --opts cover="le cover.png" --tags ...'''
options={
    'cover':''
}

def update_config(opt):
    global options
    options = {**options,**opt}

def download_video(res) -> DownloadResult:
    results = DownloadResult()
    if not os.path.isfile(res) and not os.path.isdir(res) or (options['cover'] and os.path.isfile(options['cover'])):
        raise FileNotFoundError
    def append(res):        
        with DownloadResult() as result:        
            result.video_path = res
            result.cover_path = options['cover']            
            result.original = True    
            result.title = os.path.basename(res)
            result.description = '[automated upload of file %s]' % res        
        results.results.append(result)
    if os.path.isfile(res):append(res)
    else:
        for f in os.listdir(res):
            append(os.path.join(res,f))
    results.title = os.path.basename(res)   
    results.description = '[automated upload of file %s]' % res
    return results


# -*- coding: utf-8 -*-
import os
import argparse
import logging
from pathlib import Path
from .. import providers

class AttribuitedDict(dict):
    def __getattr__(self,name):
        return self[name]
pbar = None
global_args = {            
    'cookies' : {'help':'登陆： 使用 Cookies 登陆，即使用 Web API （不可多 P 上传） ( 需要 SESSDATA 及 bili_jct ) e.g.SESSDATA=cb0..; bili_jct=6750...'},
    'sms' : {'help':'登陆：使用短信验证码登陆，即使用 上传助手 API （可多 P 上传）（需手动交互）（有日获取限制，请配合 --save 使用）','default':False,'action':'store_true'},
    'load' : {'help':'登陆：使用保存过的凭据登陆，由该参数读入'},
    'save' : {'help':'登陆：向stdout输出当前登陆凭据并退出（其他输出转移至stderr）','default':False,'action':'store_true'},
    'http' : {'help':'强制使用 HTTP （不推荐）','default':False,'action':'store_true'},
    'noenv' : {'help':'上传时，不采用环境变量（如代理）','default':False,'action':'store_true'},
    'cdn'  : {'help':'上传用 CDN （限 Web API) （对应 网宿（适合海外），七牛，百度（默认），七牛，谷歌，百度）','choices':['ws','qn','bda2','kodo','gcs','bos'],'default':'bda2'},    
}
local_args = {
    'opts':{'help':'解析可选参数 ，详见 --opts 格式','default':''},
    'thread_id': {'help':'分区 ID','default':17},
    'tags': {'help':'标签','default':'转载'},
    'desc':{'help':'描述格式 e.g. "原描述：{desc}" (其他变量详见下文)（仅稿件有描述）','default':'{desc}'},
    'title':{'help':'标题格式 e.g. "[Youtube] {title} (其他变量详见下文)（使用于稿件及分P）"','default':'{title}'},    
    'seperate_parts':{'help':'不分P （e.g. --youtube [播放列表],--localfile [文件夹]）独立投稿（不分P）（Web上传默认不分 P）','default':False,'action':'store_true'},
    'no_upload':{'help':'只下载资源','default':False,'action':'store_true'},
    'original':{'help':'设置稿件为原创','default':False,'action':'store_true'},
    'no_reprint':{'help':'设置稿件不允许转载','default':False,'action':'store_true'}
}
arg_epilog = '''
变量：
    {title},{desc} 等变量适用于：
        title, desc, tags

本工具支持将给定视频源转载至哔哩哔哩

详见项目 README 以获取更多例程 ： github.com/greats3an/bilibili-toolman
'''
def setup_logging():
    import coloredlogs
    coloredlogs.DEFAULT_LOG_FORMAT = '[ %(asctime)s %(name)8s %(levelname)6s ] %(message)s'
    coloredlogs.install(0)
    logging.getLogger('urllib3').setLevel(logging.CRITICAL)
    logging.getLogger('PIL.Image').setLevel(logging.CRITICAL)

def prepare_temp(temp_path : str):    
    if not os.path.isdir(temp_path):os.mkdir(temp_path)
    os.chdir(temp_path)    
    return True

def report_progress(current, max_val): 
    from . import precentage_progress
    precentage_progress.report(current,max_val)

def _enumerate_providers():
    provider_dict = dict()
    for provider in dir(providers):
        if not 'provider_' in provider:
            continue
        provider_name = provider.replace('provider_', '')
        provider_dict[provider_name] = getattr(providers, provider)
    return provider_dict

provider_args = _enumerate_providers()

def _create_argparser():
    p = argparse.ArgumentParser(description='使用帮助',formatter_class=argparse.RawTextHelpFormatter,epilog=arg_epilog)
    g = p.add_argument_group('身份设置 （随方式优先级排序）')
    for arg_key, arg in global_args.items():
        g.add_argument('--%s' % arg_key, **arg)
    # global args
    g = p.add_argument_group('上传设置')
    for arg_key, arg in local_args.items():                
        g.add_argument('--%s' % arg_key, **arg)
    # local args (per source)
    g = p.add_argument_group('解析可选参数 "opts" （格式 ： [参数1]=[值1];[参数2]=[值2] (query-string)）')
    for provider_name, provider in provider_args.items():
        g.add_argument(
            '--%s' % provider_name, 
            metavar='%s-URL' % provider_name.upper(),
            type=str, help='%s\n   参数:%s'%(provider.__desc__,provider.__cfg_help__)
        )
    return p

def sanitize(string):
    # limits characters to CJK & ascii chars
    import re
    return re.sub('[^\u0000-\u007F\uac00-\ud7a3\u3040-\u30ff\u4e00-\u9FFF]*','',string) # remove emojis

def truncate(string,max):
    # truncate & add ... to string over the length of `max`
    if len(string) > max:string = string[:max-3] + '...'
    return string

def prase_args(args: list):    
    if len(args) < 2:
        parser = _create_argparser()    
        parser.print_help()
        return
    args.pop(0)  # remove filename
    parser = _create_argparser()    
    global_args_dict = AttribuitedDict()
    for k, v in parser.parse_args(args).__dict__.items():
        if k in global_args:
            global_args_dict[k] = v
            if not '--%s' % k in args:continue
    '''pre-parse : fetch global args,then remove them'''
    local_args_group = []
    current_line = []
    current_provider = ''

    def add(): 
        args = parser.parse_args(current_line).__dict__       
        args['resource'] = args[current_provider]
        local_args_group.append((provider_args[current_provider],AttribuitedDict(args)))
    i = 0
    while i < len(args):
        if args[i][2:] in provider_args:
            # a new proivder. terminates till next provider
            current_provider = args[i][2:]
            current_line = ['--' + current_provider]
            for i in range(i + 1,len(args)):
                if args[i][2:] in provider_args: break
                current_line.append(args[i])
            add()            
            i -= 1
        i+=1            
    return global_args_dict, local_args_group


# -*- coding: utf-8 -*-
'''bilibili - PC API implementation'''
from time import time
from typing import Iterable, Optional, Text, Tuple, Union
from requests.sessions import Session
from urllib.parse import quote,urlencode
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5
from hashlib import md5
from base64 import b64encode
import math

from .web import BiliSession as WebBiliSession,logger
from .common import FileIterator, JSONResponse, LoginException , ReprExDict, file_manager, check_file 
from .common.submission import Submission

def PCOnlyAPI(_classmethod):
    def wrapper(self,*a,**k):
        assert type(self) == BiliSession,"限 PC API 使用"
        return _classmethod(self,*a,**k)
    return wrapper     

class Crypto:
    '''THANK YOU! https://github.com/FortuneDayssss/BilibiliUploader'''
    APPKEY = '661dbf0ee792f083'
    APPSECRET = '397c810e126bfbbb584667583e11976a'
    # thx x69dbg ( ͡° ͜ʖ ͡°)
    @staticmethod
    def iterable_md5(stream : Iterable) -> str:
        md5_ = md5()
        for chunk in stream:
            md5_.update(chunk)
        return md5_.hexdigest()            

    @staticmethod
    def md5(data : Union[str,bytes]) -> str:
        '''generates md5 hex dump of `str` or `bytes`'''
        if type(data) == str:
            return md5(data.encode()).hexdigest()
        return md5(data).hexdigest()

    @staticmethod
    def rsa_pkcs115_encrypt_as_base64(content: bytes, pkcs115_pubkey : str) -> str:
        key = RSA.import_key(pkcs115_pubkey)
        cipher = PKCS1_v1_5.new(key)
        return b64encode(cipher.encrypt(content))

    @staticmethod
    def encrypt_login_password(password: str, oauth2_getkey_json: dict) -> str:
        _hash, key = oauth2_getkey_json['data']['hash'], oauth2_getkey_json['data']['key']
        return Crypto.rsa_pkcs115_encrypt_as_base64((_hash + password).encode('utf-8'), key)

    @staticmethod
    def sign(data : Union[str,dict]) -> str: 
        '''salted sign funtion for `dict`(converts to qs then parse) & `str`'''
        if isinstance(data,dict):
            _str = urlencode(data)        
        elif type(data) != str:
            raise TypeError
        return Crypto.md5(_str + Crypto.APPSECRET)

class SingableDict(dict):
    @property
    def sorted(self):
        '''returns a alphabetically sorted version of `self`'''
        return dict(sorted(self.items()))
    @property
    def signed(self):
        '''returns our sorted self with calculated `sign` as a new key-value pair at the end'''
        _sorted = self.sorted
        return {**_sorted, 'sign': Crypto.sign(_sorted)}

class ClientUploadChunk(FileIterator):
    url_endpoint : str
    params : dict
    headers : dict    
    files : dict
    cookies : dict
    session : Session        

    def upload_via_session(self,session = None):
        chunk_bytes = self.to_bytes()
        for retries in range(1,BiliSession.RETRIES_UPLOAD_ID+1):
            try:
                (session or self.session).post(
                    self.url_endpoint,
                    params=self.params,
                    headers=self.headers,
                    files={
                        **self.files,
                        'md5':(None,Crypto.md5(chunk_bytes)),
                        'file': (self.path, chunk_bytes, 'application/octet-stream')
                    },
                    cookies=self.cookies,            
                )
                return True
            except Exception as e:
                logger.warning('第 %s 次重试时：%s' % (retries,e))

class BiliSession(WebBiliSession):
    '''哔哩哔哩上传助手 API'''
    # this part reused A LOT of old code, some can still be replaced with thier PC counterparts 
    UPLOAD_CHUNK_SIZE = 2 * (1 << 20)    
    BUILD_VER = (2, 3, 0,1073)
    BUILD_NO = int(BUILD_VER[0] * 1e8 + BUILD_VER[1] * 1e6 + BUILD_VER[2] * 1e2 + BUILD_VER[3])
    BUILD_STR = '.'.join(map(lambda v: str(v), BUILD_VER))    

    UPLOAD_PROFILE = 'ugcfr/pc3'
    
    DEFAULT_UA = ''
    
    RETRIES_UPLOAD_ID = 5
    
    MISC_MAX_TITLE_LENGTH = 80
    MISC_MAX_DESCRIPTION_LENGTH = 1500 # somehow larger than expected

    def __init__(self) -> None:
        Session.__init__(self) # no need to init the web variant
        self.headers = {
            'User-Agent': BiliSession.DEFAULT_UA,
            'Accept-Encoding': 'gzip,deflate',
        }
        self.login_tokens = dict()
    # region Properties
    @property
    def mid(self):
        return self.login_tokens['mid']

    @property
    def access_token(self):
        return self.login_tokens['access_token']

    @property
    def access_key_param(self):
        '''Singed dictionary containing only `access_key` key-value pair'''
        return SingableDict({'access_key':self.access_token}).signed
        
    # endregion

    # region Overrides
    def _preupload(self):
        return self.get(
            "http://member.bilibili.com/preupload",
            params={
                'access_key':self.access_token,
                'mid':self.mid,
                'profile': self.UPLOAD_PROFILE
            }
        )

    def _upload_cover(self,image_binary : bytes,image_mime : str):
        return self.post(
            "https://member.bilibili.com/x/vu/client/cover/up",
            params=self.access_key_param,
            files={'file':('cover.png',image_binary,image_mime)}
        )      

    def _submit_submission(self, submission : Submission):
        return self.post(
            "https://member.bilibili.com/x/vu/client/add",
            params=self.access_key_param,                                    
            json={'build': self.BUILD_VER[-1],**submission.archive}
        )        


    def _list_archives(self,params):
        return self.get(
            'https://member.bilibili.com/x/client/archives',
            params=SingableDict({
                "access_key": self.access_token,
                **params
            }).signed
        )

    @JSONResponse
    def _view_archive(self, bvid):
        return self.get(
            'https://member.bilibili.com/x/client/archive/view',
            params=SingableDict({
                "access_key": self.access_token,
                "bvid": bvid,
                'build': self.BUILD_VER[-1]
            }).signed
        )

    def _edit_archive(self, json: dict):                
        return self.post(
            "https://member.bilibili.com/x/vu/client/edit",
            params=self.access_key_param,
            json=SingableDict({
                'build': self.BUILD_VER[-1],
                'no_reprint':0,
                'open_elec':0,
                **json,                
            }).sorted 
        )

    def _delete_archive(self,bvid):
        return self.post(
            "http://member.bilibili.com/x/client/archive/delete",
            params=self.access_key_param,
            data={'bvid':bvid}
        )
    # endregion

    # region Client-specific APIs
    @JSONResponse
    def _oauth2_getkey(self):
        return self.get("https://passport.bilibili.com/x/passport-login/web/key", params=SingableDict({
            'appkey': Crypto.APPKEY,
            'platform':'pc'            
        }).signed)

    def _post_complete_upload(self,complete_url,size,basename,md5,chunkcount):
        return self.post(
            complete_url,
            data={
                'chunks': chunkcount,
                'filesize': size,
                'md5': md5,
                'name': basename,
                'version': self.BUILD_STR,
            })

    @PCOnlyAPI
    @JSONResponse
    def DeleteArchive(self,bvid):
        '''删除作品

        Args:
            bvid
        '''
        return self._delete_archive(bvid)

    @PCOnlyAPI
    @JSONResponse
    def LoginViaUsername(self, username: str, password: str):
        '''用户名密码登陆

        Args:
            username (str), password (str)

        Returns:
            dict
        '''
        raise DeprecationWarning("Deprecated by %s. Use SMS login instead."  % self.BUILD_STR)
        oauth_json = self._oauth2_getkey()
        resp = self.post(
            "https://passport.bilibili.com/x/passport-login/oauth2/login",
            data=SingableDict({
                'appkey': Crypto.APPKEY,
                'platform': "pc",                
                'password': Crypto.encrypt_login_password(password, oauth_json),                                
                'username': quote(username),
                'ts': int(time()*1000),
                'device_name':'',
                'device_id':'',
                'buvid':''
        }).signed)
        try:
            self.login_tokens.update(resp.json()['data']['token_info'])
        except Exception as e:
            raise LoginException(resp,e)
        return resp
    
    @PCOnlyAPI
    @JSONResponse
    def RenewSMSCaptcha(self,tel : str,cid = 86):
        '''发送验证码

        Args:
            tel (str) : 手机号
            cid (int) : 国家代码

        Returns:
            dict
        '''        
        resp = self.post(
            "https://passport.bilibili.com/x/passport-login/sms/send",
            data=SingableDict({
                'appkey': Crypto.APPKEY,
                'platform': "pc",                
                'tel':str(tel),
                'cid':str(cid),
                'ts': int(time()*1000),
                'device_name':'',
                'device_id':'',
                'buvid':''
        }).signed)
        try:
            self.login_tokens['captcha_key'] = resp.json()['data']['captcha_key']
            assert self.login_tokens['captcha_key']               
        except Exception as e:
            raise LoginException(resp,e)        
        return resp

    @PCOnlyAPI
    @JSONResponse
    def LoginViaSMSCaptcha(self,tel : str,code : int,cid = 86):        
        '''验证码登陆

        Args:                    
            tel (str) : 手机号
            code (str) : 验证码
            cid (int) : 国家代码

        Returns:
            dict
        '''
        assert self.login_tokens.get('captcha_key',None),"`RenewSMSCaptcha` not called or failed."
        resp = self.post(
            "https://passport.bilibili.com/x/passport-login/login/sms",
            data=SingableDict({
                'appkey': Crypto.APPKEY,
                'platform': "pc",                
                'captcha_key' : self.login_tokens["captcha_key"],
                'code' : str(code),
                'tel':str(tel),
                'cid':str(cid),                
                'ts': int(time()*1000),
                'device_name':'',
                'device_id':'',
                'buvid':''
        }).signed)
        try:
            self.login_tokens.update(resp.json()['data']['token_info'])
        except Exception as e:
            raise LoginException(resp,e)
        return resp       
    
    def UploadVideo(self, path: str) -> Tuple[str,None]:
        '''上传视频

        Args:
            path (str): 视频文件路径

        Returns:
            Tuple[str,None]: [远端 URI,None]        
        '''
        path, basename, size = check_file(path)        
        preupload_token = self._preupload().json()
        # preprae the chunks then uploads them
        chunksize = self.UPLOAD_CHUNK_SIZE
        chunkcount = math.ceil(size/chunksize)
        file_manager.open(path)
        logger.debug('上传分块: %s' % chunkcount)
        logger.debug('分块大小: %s B' % chunksize)
        def iter_chunks():
            for chunk_n in range(0,chunkcount):
                start = chunksize * chunk_n
                chunk = ClientUploadChunk(path,start,min(start + chunksize,size))
                chunk.session = self
                chunk.url_endpoint = preupload_token['url']
                chunk.files = {
                    'version': (None, self.BUILD_STR),
                    'filesize': (None, chunksize),
                    'chunk': (None, chunk_n),
                    'chunks': (None, chunkcount),                               
                }
                chunk.cookies = {
                    'PHPSESSID':preupload_token['filename']
                }
                yield chunk              
        self._upload_chunks_to_endpoint_blocking(iter_chunks())        
        # recalulating md5        
        md5_ = Crypto.iterable_md5(FileIterator(path,0,size))        
        file_manager.close(path)
        logger.debug('MD5: %s' % md5_)
        # finalizing upload
        post_r = self._post_complete_upload(preupload_token['complete'],size,basename,md5_,chunkcount)
        logger.debug('上传完毕: %s' % ReprExDict(post_r.json()))
        return preupload_token['filename'], None
    # endregion

    # region Pickling
    def __dict__(self):
        return {'cookies':self.cookies,'login_tokens':self.login_tokens,'session':self}
    
    def update(self,state_dict : dict):
        self.cookies = state_dict['cookies']        
        self.login_tokens = state_dict['login_tokens']
    # endregion



# -*- coding: utf-8 -*-
# region Setup
from bilibili_toolman.bilisession.common import LoginException
from ..bilisession.web import BiliSession
from ..bilisession import logger
from ..bilisession.common.submission import Submission
from ..providers import DownloadResult
from . import AttribuitedDict,setup_logging,prepare_temp,prase_args,sanitize,truncate,local_args as largs
from collections import defaultdict
import base64,logging,sys,urllib.parse

TEMP_PATH = 'temp'
sess : BiliSession

def download_sources(provider,arg) -> DownloadResult:
    resource = arg.resource
    opts = arg.opts
    try:
        opts = urllib.parse.parse_qs(opts,keep_blank_values=True)
        provider.update_config({k:v[-1].replace(';','') for k,v in opts.items()}) # shouldnt you use & anyways?
    except:opts = '无效选项'
    '''Passing options'''
    logger.info('下载源视频')
    logger.info('  - Type: %s - %s' % (provider.__name__,provider.__desc__))
    logger.info('  - URI : %s' % resource)    
    '''Downloading source'''
    try:
        return provider.download_video(resource)
    except Exception as e:
        logger.error('无法下载指定资源 - %s' % e)
        return        

def upload_sources(sources : DownloadResult,arg):
    '''To perform a indivudial task

    If multiple videos are given by the provider,the submission will be in multi-parts (P)
    Otherwise,only the given video is uploaded as a single part subject

    Args:

        provider - one of the modules of `providers`
        arg - task arguments dictionary    
            * resource - resoucre URI (must have)
            - opts     - options for uploader in query string e.g. format=best            
            - See `utils.local_args` for more arguments,along with thier details        
    '''    
    submission = Submission()    
    if not sources:return None,True
    logging.info('上传资源数：%s' % len(sources.results))    
    def format(source):
        blocks      = defaultdict(lambda:'...',{'title':source.title,'desc':source.description,**source.extra})
        title       = truncate(sanitize(arg.title.format_map(blocks)),sess.MISC_MAX_TITLE_LENGTH)   
        description = truncate(sanitize(arg.desc .format_map(blocks)),sess.MISC_MAX_DESCRIPTION_LENGTH)        
        return blocks,title,description        
            
    for source in sources.results:       
        '''If one or multipule sources'''        
        blocks,title,description = format(source)
        logger.info('准备上传: %s' % title)
        '''Summary trimming'''
        endpoint = None
        for _ in range(0,sess.RETRIES_UPLOAD_ID):
            try:
                endpoint, bid = sess.UploadVideo(source.video_path)
                cover_url = sess.UploadCover(source.cover_path)['data']['url'] if source.cover_path else ''
                break
            except Exception as e:
                logger.warning('%s 上传失败! - %s - 重试' % (source,e))                
        if not endpoint:
            logger.error('URI 获取失败 - 跳过')
            continue
        logger.info('资源已上传')
        from bilibili_toolman.cli import precentage_progress
        precentage_progress.close()
        with Submission() as video:
            '''Creatating a video per submission'''
            video.cover_url = cover_url
            video.video_endpoint = endpoint
            video.biz_id = bid
            '''Sources identifiers'''   
            if not arg.original:
                video.copyright = Submission.COPYRIGHT_REUPLOAD
                video.source = sources.soruce         
            else:
                video.copyright = Submission.COPYRIGHT_ORIGINAL
                video.source = ''
            video.thread = arg.thread_id
            video.tags = arg.tags.format_map(blocks).split(',')
            video.description = description # This don't gets shown per video. Only the main description is shown
            video.title = title # This shows up as title per-part
        '''Use the last given thread per multiple uploads,while the tags are extended.'''
        submission.thread = video.thread or submission.thread        
        submission.tags.extend(video.tags)
        submission.videos.append(video) # to the main submission
    '''Filling submission info'''        
    submission.copyright = submission.COPYRIGHT_REUPLOAD if not arg.original else submission.COPYRIGHT_ORIGINAL
    submission.no_reprint = submission.REPRINT_ALLOWED if not arg.no_reprint else submission.REPRINT_DISALLOWED
    '''Use title & description from main sources (e.g. Uploading a YT Playlist)'''
    blocks,title,description = format(sources)
    submission.title = title
    submission.description = description # This is the only description that gets shown
    submission.source = sources.soruce
    '''Upload cover images for all our submissions as well'''
    cover_url = sess.UploadCover(sources.cover_path)['data']['url'] if sources.cover_path else ''
    submission.cover_url = cover_url    
    '''Finally submitting the video'''
    submit_result=sess.SubmitSubmission(submission,seperate_parts=arg.seperate_parts)  
    dirty = False
    for result in submit_result['results']:
        if result['code'] == 0:logger.info('上传成功 - BVid: %s' % result['data']['bvid'])        
        else:
            logger.warning('%s 上传失败 : %s' % (submission,result['message']))
            dirty = True
    return submit_result,dirty

global_args,local_args = None,None

def setup_session():
    '''Setup session with cookies in query strings & setup temp root'''
    global sess
    def setup_params(sess):
        if global_args.http:
            logger.warning('强制使用 HTTP')
            sess.FORCE_HTTP = True

        if global_args.noenv:
            logger.warning('不使用环境变量；请求将绕过代理')
            sess.trust_env = False                    
    if global_args.cookies:
        from ..bilisession.web import BiliSession
        sess = BiliSession()    
        setup_params(sess)
        sess.LoginViaCookiesQueryString(global_args.cookies)
        self_ = sess.Self
        if not 'uname' in self_['data']:
            logger.error('Cookies无效: %s' % self_['message'])        
            return False        
        # for task,arg in local_args:
        #     arg['seperate_parts'] = True
        logger.warning('Web端 API 需 Lv3+ 及 1000+ 关注量才可多 P 上传，若出错请悉知')
        return True
    elif global_args.sms:        
        from ..bilisession.client import BiliSession
        sess = BiliSession()
        setup_params(sess)
        logger.warning('短信验证码有日发送条数限制（5 条），无论验证成败，超限后将无法发送验证码')
        logger.warning('建议使用 --save 保存验证凭据，再次使用则利用 --load 读取，而不需再次登陆')
        logger.info('准备登陆，输入手机号后，按Enter发送验证码')
        phone = input()                
        sess.RenewSMSCaptcha(phone)
        logger.debug('验证码已发送')
        for i in range(0,5):
            logger.info('输入验证码:')
            captcha=input() 
            try:
                sess.LoginViaSMSCaptcha(phone,captcha)                
                return True                
            except LoginException as e:
                logger.error('验证码无效，请重试：%s' % e)                
        return False
        pass
    elif global_args.load:    
        from ..bilisession.web import BiliSession
        sess = BiliSession.from_bytes(base64.b64decode(global_args.load))
        setup_params(sess)
        logger.info('加载之前的登陆凭据')
        return True
    else:
        logger.error('缺失认证信息')
        return False

def __main__():
    global global_args,local_args
    setup_logging()
    args = prase_args(sys.argv)
    if args:
        global_args,local_args = args
    else:
        sys.exit(1)
    '''Parsing args'''
    if not setup_session():
        logger.fatal('登陆失败！')
        sys.exit(2)            
    
    if global_args.save:
        logger.info('保存登陆凭据')        
        print(base64.b64encode(sess.to_bytes()).decode())
        sys.exit(0)

    if sess.DEFAULT_UA: # using Web APIs
        bup = {'ws','qn','bda2'}
        bupfetch = {'kodo','gcs','bos'}
        if global_args.cdn in bup:
            sess.UPLOAD_PROFILE = 'ugcupos/bup'                                
        elif global_args.cdn in bupfetch:
            sess.UPLOAD_PROFILE = 'ugcupos/bupfetch'
        sess.UPLOAD_CDN     = global_args.cdn
        logger.info('CDN ： %s [%s]' % (sess.UPLOAD_CDN,sess.UPLOAD_PROFILE))            
        logger.warning('Web 端 API @ ID:%s' % sess.Self['data']['uname'])
    else: # using client APIs
        try:
            logger.warning('上传助手 API @ MID:%s' % sess.mid)    
        except:
            logger.fatal('无效登陆凭据')
            sys.exit(1)
    prepare_temp(TEMP_PATH)    
    # Output current settings        
    logger.info('任务总数: %s' % len(local_args))        
    success,failure = [],[]
    fmt = lambda s: ('×','√')[s] if type(s) is bool else s
    for provider,arg_ in local_args:
        arg = AttribuitedDict(arg_)
        logger.info('任务信息：')
        for k,v in largs.items():
            logger.info('  - %s : %s' % (list(v.values())[0].split()[0],fmt(arg[k])))
        sources = download_sources(provider,arg)                   
        if arg.no_upload:
            logger.warn('已跳过上传')
        else:
            result,dirty = upload_sources(sources,arg)
            if not dirty:success.append((arg,result))
            else: failure.append((arg,None))
          
    if not failure:
        logger.info('任务完毕')
        sys.exit(0)
    logger.warning('上传未完毕')
    sys.exit(1)


# -*- coding: utf-8 -*-
'''bilibili - Web API implmentation'''
from concurrent.futures.thread import ThreadPoolExecutor
import json,pickle,gzip
from threading import Thread
from requests import Session
from typing import List, Tuple
from . import logger
import math,time,mimetypes,base64

from .common import JSONResponse , FileIterator, ReprExDict , file_manager , chunk_queue , check_file
from .common.submission import Submission , create_submission_by_arc

def WebOnlyAPI(_classmethod):
    def wrapper(self,*a,**k):
        assert type(self) == BiliSession,"限 Web API 使用"
        return _classmethod(self,*a,**k)
    return wrapper     

class WebUploadChunk(FileIterator):
    url_endpoint : str
    params : dict
    headers : dict    
    session : Session

    
    def upload_via_session(self,session = None):
        for retries in range(1,BiliSession.RETRIES_UPLOAD_ID+1):
            try:
                (session or self.session).put(
                    self.url_endpoint,
                    params=self.params,
                    headers=self.headers,
                    data=self
                )
                return True
            except Exception as e:
                logger.warning('第 %s 次重试时：%s' % (retries,e))

class BiliSession(Session):
    '''哔哩哔哩网页上传 API'''
    BUILD_VER = (2, 8, 12)
    BUILD_NO = int(BUILD_VER[0] * 1e6 + BUILD_VER[1] * 1e4 + BUILD_VER[2] * 1e2)
    BUILD_STR = '.'.join(map(lambda v: str(v), BUILD_VER))
    '''Build variant & version'''
    
    DEFAULT_UA = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36'
    
    UPLOAD_PROFILE = 'ugcupos/bup'    
    UPLOAD_CDN     = 'bda2'

    RETRIES_UPLOAD_ID = 5

    DELAY_FETCH_UPLOAD_ID = .1
    DELAY_RETRY_UPLOAD_ID = 1
    DELAY_VIDEO_SUBMISSION = 30
    DELAY_REPORT_PROGRESS = 1

    WORKERS_UPLOAD = 3

    MISC_MAX_TITLE_LENGTH = 80
    MISC_MAX_DESCRIPTION_LENGTH = 2000

    FORCE_HTTP = False   

    def request(self, method: str, url,*a,**k):
        if self.FORCE_HTTP and url[:5] == 'https':
            url='http' + url[5:]                                   
        return super().request(method, url,*a,**k)

    def __init__(self, cookies: str = '') -> None:      
        Session.__init__(self)
        self.LoginViaCookiesQueryString(cookies)
        self.headers['User-Agent'] = self.DEFAULT_UA       
        
    # region Web-client APIs    
    @WebOnlyAPI
    def LoginViaCookiesQueryString(self,cookies:str):
        '''设置本 Session 的 Cookies

        Args:
            cookies (str): e.g. SESSDATA=cb0..; bili_jct=6750...
        '''
        if not cookies: return
        for item in cookies.replace(' ', '').split(';'):
            if '=' in item:
                self.cookies.set(*item.split('=')[:2])
            else:
                self.cookies.set(item,'')
        return True    

    def _self(self):
        return self.get('https://api.bilibili.com/x/web-interface/nav')

    @property
    @JSONResponse
    def Self(self):
        '''个人信息，限网页端使用'''
        return self._self()

    @JSONResponse
    def _upload_status(self, endpoint, name, upload_id, biz_id):
        '''检查网页端上传结果，限网页端使用'''
        return self.post(endpoint, params={
            'output': 'json',
            'profile': self.UPLOAD_PROFILE,
            'name': name,
            'uploadId': upload_id,
            'biz_id': biz_id
        })

    def _list_archives(self,params):
        return self.get('https://member.bilibili.com/x/web/archives',params=params)

    @JSONResponse
    def ListArchives(self,pubing=True,pubed=True,not_pubed=True,pn=1,ps=10):
        '''分页查看已上传的作品,*推荐使用`ListSubmissions`*

        Args:
            pubing (bool, optional): 是否获取*正在审核*的作品. Defaults to True.
            pubed (bool, optional): 是否获取*已发布*的作品. Defaults to True.
            not_pubed (bool, optional): 是否获取*被打回*的作品. Defaults to True.
            pn (int, optional): 页码. Defaults to 1.
            ps (int, optional): 个数. Defaults to 10.

        Returns:
            dict
        '''
        return self._list_archives({
            'status':('%s%s%s' % (',is_pubing'*pubing,',pubed'*pubed,',not_pubed'*not_pubed))[1:],
            'pn':pn,
            'ps':ps,
            'interactive':1,
            'coop':1
        })
    
    @JSONResponse
    def _view_archive(self,bvid):
        return self.get('https://member.bilibili.com/x/web/archive/view',params={'bvid':bvid})
    
    def _edit_archive(self,json : dict):
        return self.post('https://member.bilibili.com/x/vu/web/edit',json=json,params={'csrf': self.cookies.get('bili_jct')})

    @JSONResponse
    def ViewPublicArchive(self,bvid):
        '''以 BVID 获取公布作品信息

        Args:
            bvid    
        仅适用于已发布的作品
        '''
        return self.get('https://api.bilibili.com/x/web-interface/view',params={'bvid':bvid})
        
    @JSONResponse
    def ViewPlayerArchive(self,cid : int,bvid : str):
        '''获取作品中子视频详情 （含字幕相关字段）

        Args:
            cid (int): cid
            bvid (str): bvid
        '''
        return self.get('https://api.bilibili.com/x/player/v2',params={'cid':cid,'bvid':bvid})

    @JSONResponse
    def EditSubmission(self,submission : Submission):
        '''编辑作品，适用于重新上传

        Args:
            submission (Submission): 可由 `ViewArchive` 取得

        Returns:
            dict
        '''
        return self._edit_archive({
                "aid": submission.aid,
                "copyright": submission.copyright,
                "videos": submission.videos.archive,
                "source": submission.source,
                "tid": int(submission.thread),
                "cover": submission.cover_url,
                "title": submission.title,
                "tag": ','.join(set(submission.tags)),
                "desc_format_id": 31,
                "desc": submission.description, 
        })

    def ViewSubmission(self,bvid) -> Submission:
        '''以 BVid 获取作品信息

        Args:
            bvid

        Returns:
            Submission: 作品信息
        '''
        arc = self._view_archive(bvid)['data']
        return create_submission_by_arc(arc)

    def ListSubmissions(self,pubing=True,pubed=True,not_pubed=True,limit=1000) -> List[Submission]:
        '''查看已上传的作品

        Args:
            pubing (bool, optional): 是否获取*正在审核*的作品. Defaults to True.
            pubed (bool, optional): 是否获取*已发布*的作品. Defaults to True.
            not_pubed (bool, optional): 是否获取*被打回*的作品. Defaults to True.
            limit (int, optional): 最多获取量. Defaults to 1000.

        Raises:
            Exception: 被限流时引发

        Returns:
            List[Submission]: 请求到的作品

        注：此 API 无法获取完整作品信息，推荐通过所得BVID以其他API检索
        '''
        args = pubing,pubed,not_pubed
        submissions = []
        count = 0
        def add_to_submissions(arcs):       
            nonlocal count
            for arc in arcs['arc_audits']:
                count += submissions.append(create_submission_by_arc(arc)) or 1
                if count >= limit:return False  
            return True             
        arc = self.ListArchives(*args,pn=1)['data']
        result = add_to_submissions(arc)
        if result:
            for pn in range(2,math.ceil(arc['page']['count'] / arc['page']['ps']) + 1):
                add_to_submissions(self.ListArchives(*args,pn=pn)['data'])
        return submissions

    def _preupload(self, name='a.flv', size=0):        
        return self.get('https://member.bilibili.com/preupload', params={
            'name': name,
            'size': int(size),
            'r': 'upos',
            'profile': self.UPLOAD_PROFILE,
            'ssl': 0,
            'version': self.BUILD_STR,
            'build': self.BUILD_NO,
            'upcdn': self.UPLOAD_CDN,
            'probe_version': self.BUILD_NO
        })

    def _upload_id(self, endpoint):
        time.sleep(self.DELAY_FETCH_UPLOAD_ID)  # adding delay as the `auth` token needs to be updated server-side
        return self.post(endpoint + '?uploads', params={
            'output': 'json'
        },headers={
            'Origin': 'https://member.bilibili.com',
            'Referer':'https://member.bilibili.com/'
        })    

    def _upload_chunks_to_endpoint_blocking(self,chunk_iter : List[WebUploadChunk]):
        '''consuming all chunks through any means,blocks code until done'''
        for chunk in chunk_iter:chunk_queue.put(chunk)
        executor = ThreadPoolExecutor(max_workers=self.WORKERS_UPLOAD)
        class ConsumerThread(Thread):
            daemon = True
            def run(self) -> None:
                while chunk_queue.unfinished_tasks >= 0:
                    chunk : WebUploadChunk = chunk_queue.get()             
                    future = executor.submit(chunk.upload_via_session)           
                    future.add_done_callback(lambda future:chunk_queue.task_done())
        tConsume = ConsumerThread()
        tConsume.start()
        from . import cli        
        while chunk_queue.unfinished_tasks >= 0:       
            read_all,size_all = 0 , 0
            for v in file_manager.values():
                read_all += v['read'] 
                size_all += v['length']
            cli.report_progress(read_all,size_all)
            if chunk_queue.unfinished_tasks == 0: break
            time.sleep(self.DELAY_REPORT_PROGRESS)
        del tConsume
        return True

    def UploadVideo(self, path: str) -> Tuple[str,int]:
        '''上传视频

        Args:
            path (str): 视频文件路径

        Returns:
            Tuple[str,str]: [远端 URI,biz_id]        
        '''
        path, basename, size = check_file(path)
        def generate_upload_chunks(name, size):
            def fetch_upload_id():
                '''Generating uplaod chunks'''            
                for i in range(1,self.RETRIES_UPLOAD_ID + 1):
                    try:
                        config = self._preupload(name=name, size=size).json()
                        self.headers['X-Upos-Auth'] = config['auth']
                        '''X-Upos-Auth header'''
                        endpoint = 'https:%s/ugcboss/%s' % (config['endpoint'], config['upos_uri'].split('/')[-1])
                        logger.debug('远端结点： %s' % endpoint)
                        logger.debug('第 %s 次刷新 TOKEN...' % i)
                        upload_id = self._upload_id(endpoint).json()['upload_id']
                        return config,endpoint,upload_id
                    except Exception as e:               
                        logger.warning('第 %s 上传未成功,重试...' % i)
                        time.sleep(self.DELAY_RETRY_UPLOAD_ID)            
                return None,None,None
            config,endpoint,upload_id = fetch_upload_id()
            if not upload_id:
                raise Exception("经 %s 次重试后仍无法获取 TOKEN" % self.RETRIES_UPLOAD_ID)
            '''Upload endpoint & keys'''
            chunksize = config['chunk_size']
            chunkcount = math.ceil(size/chunksize)
            file_manager.open(path)
            logger.debug('上传分块: %s' % chunkcount)
            logger.debug('分块大小: %s B' % chunksize)
            def iter_chunks():
                for chunk_n in range(0,chunkcount):   
                    start = chunksize * chunk_n
                    end = min(start + chunksize,size)
                    chunk = WebUploadChunk(path,start,end)
                    chunk.url_endpoint = endpoint
                    chunk.session = self
                    chunk.params = {
                        'partNumber': chunk_n + 1,
                        'uploadId': upload_id,
                        'chunk': chunk_n,
                        'chunks': chunkcount,
                        'start': start,
                        'end': end,
                        'total': size
                    }
                    chunk.headers = {
                        'X-Upos-Auth': config['auth']
                    }
                    yield chunk            
            config['upload_id'] = upload_id
            return endpoint, config, iter_chunks()
        endpoint, config, chunks = generate_upload_chunks(basename, size)
        '''Generates upload config'''    
        self._upload_chunks_to_endpoint_blocking(chunks)        
        '''Wait for current upload to finish'''        
        file_manager.close(path)
        state = self._upload_status(endpoint, basename, config['upload_id'], config['biz_id'])
        if(state['OK']==1):
            logger.debug('上传完毕: %s' % ReprExDict(state))
        else:
            raise Exception('上传失败: %s' % ReprExDict(state))
        return endpoint, config['biz_id']
    
    def _upload_cover(self,image_binary : bytes,image_mime : str):
        return self.post('https://member.bilibili.com/x/vu/web/cover/up', data={
            'cover': 'data:{%s};base64,' % image_mime + base64.b64encode(image_binary).decode(),
            'csrf': self.cookies.get('bili_jct')
        })

    @JSONResponse
    def UploadCover(self, path: str):
        '''上传封面'''
        mime = mimetypes.guess_type(path)[0] or 'image/png' # fall back to png
        logger.debug('%s -> %s' % (path,mime))
        content = open(path, 'rb').read()
        logger.debug('上传封面图 (%s B)' % len(content))
        return self._upload_cover(content,mime)

    def _submit_submission(self,submission : Submission):
        return self.post(
            "https://member.bilibili.com/x/vu/web/add/v3",
            json={
                **submission.archive, 
                'csrf':self.cookies.get('bili_jct')
            },
            params={'csrf': self.cookies.get('bili_jct')}
        )

    def SubmitSubmission(self, submission: Submission,seperate_parts=False):
        '''提交作品，适用于初次上传；否则请使用 `EditSubmission`

        Args:
            submission (Submission): 作品
            seperate_parts (bool, optional): 是否将多个子视频单独上传. Defaults to False.        
        '''
        if not seperate_parts:
            logger.debug('准备提交多 P 内容: %s' % submission.title)            
            result = self._submit_submission(submission).json()
            return {'code:':result['code'],'results':[result]}
        else:
            results = []
            codes = 0
            for submission in submission.videos:
                logger.debug('准备提交单 P 内容: %s' % submission.title)
                while True:
                    result = self._submit_submission(submission).json()
                    if result['code'] == 21070:
                        logger.warning('请求受限（限流），准备重试')
                        time.sleep(self.DELAY_VIDEO_SUBMISSION)
                        continue 
                    elif result['code'] != 0:
                        logger.critical('其他错误 (%s): %s - 跳过上传' % (result['code'],result['message']))                        
                        break       
                    else:
                        break             
                codes += result['code'] # we want to see if its 0 or else
                results.append(result)
            return {'code':codes,'results':results}
    
    @WebOnlyAPI
    @JSONResponse
    def ListReceivedSubtitles(self,page=1,size=10,status=0):
        '''Web 端 - 枚举已收到的字幕

        Args:
            page (int, optional): 页码. Defaults to 1.
            size (int, optional): 单页个数. Defaults to 10.
            status (int , optional): 状态. (0=全部，2=待审核,5=已发布). Defaults to 5
        '''
        return self.get("https://api.bilibili.com/x/v2/dm/subtitle/search/assist",params={
            'type':1,'status':status,'page':page,'size':size
        })

    @WebOnlyAPI
    @JSONResponse
    def ListSubmittedSubtitles(self,page=1,size=10,status=0):
        '''Web 端 - 枚举已投稿的字幕

        Args:
            page (int, optional): 页码. Defaults to 1.
            size (int, optional): 单页个数. Defaults to 10.
            status (int , optional): 状态. (0=全部，2=待审核,5=已发布). Defaults to 5
        '''
        return self.get("https://api.bilibili.com/x/v2/dm/subtitle/search/author/list",params={
            'status':status,'page':page,'size':size
        })

    @WebOnlyAPI
    @JSONResponse
    def SaveSubtitleDraft(self,bvid : str,biz_id : int,data:dict,lang='zh-CN',submit=True):
        '''Web 端 - 提交、保存字幕

        Args:
            bvid (str) : bvid
            biz (int) : 作品 cid
            data (dict): 字幕数据 e.g. {"font_size":0.4,"font_color":"#FFFFFF","background_alpha":0.5,"background_color":"#9C27B0","Stroke":"none","body":[{"from":0,"to":5,"location":2,"content":"ts-00-05"}]}
            lang (str, optional): 语言. Defaults to 'zh-CN'.
            submit (bool, optional): 是否提交. Defaults to True.
        '''
        return self.post("https://api.bilibili.com/x/v2/dm/subtitle/draft/save",data={
            'type':1,
            'oid':biz_id,
            'lan':lang,
            'data':json.dumps(data),
            'submit':submit,
            'sign':False,
            'csrf':self.cookies.get('bili_jct'),
            'bvid':bvid
        })

    @WebOnlyAPI
    @JSONResponse
    def RevokeSubtitle(self,biz_id : int,subtitle_id:int,comment:str,type=1):
        '''Web 端 - 退回字幕

        Args:
            biz_id : 作品 cid
            subtitle_id (int): 字幕id
            comment (str): 退回理由
            type (int, optional): 退回类型，未知. Defaults to 1.    
        '''
        return self.post("https://api.bilibili.com/x/v2/dm/subtitle/assit/audit",data={
            'type':1,
            'oid':biz_id,
            'pass':False,
            'csrf':self.cookies.get('bili_jct'),
            'subtitle_id':subtitle_id,
            'reject_comment':comment
        })     

    @WebOnlyAPI
    @JSONResponse
    def GetSubtitleDetail(self,biz_id : int,subtitle_id:int):
        '''获取字幕详情

        Args:
            biz_id : 作品 cid
            subtitle_id (int): 字幕 id
        '''
        return self.get("https://api.bilibili.com/x/v2/dm/subtitle/show",params={
            'oid':biz_id,
            'subtitle_id':subtitle_id
        })
    # endregion

    # region Pickling
    def __dict__(self):
        return {'cookies':self.cookies,'session':self}
    
    def update(self,state_dict : dict):
        self.cookies = state_dict['cookies']        
    
    def to_bytes(self):                
        return gzip.compress(pickle.dumps(self.__dict__()))

    @staticmethod
    def from_bytes(b : bytes):
        unpickled = pickle.loads(gzip.decompress(b))
        sess = unpickled['session']
        sess.update(unpickled)
        return sess
    # endregion



# -*- coding: utf-8 -*-
'''Content provider modules'''
from typing import List


class DownloadResult:
    video_path : str = ''
    cover_path : str = ''

    title : str = ''
    description : str = ''    
    soruce : str = ''

    original : bool = False # now this really is just a placeholder
    results = None

    extra = dict() # BAD MOVE! though this SHOULD be overwritten and not accessed

    @property
    def results(self):
        '''A list of sub download results'''
        self._results : List[DownloadResult] # hack to allow typing of ourself
        return self._results    
    
    def __init__(self) -> None:
        self._results = []

    def __enter__(self):        
        return DownloadResult()
    def __exit__(self,*args):pass

    def __repr__(self) -> str:
        return '< title : %s , src : %s>' % (self.title,self.soruce)
from . import youtube as provider_youtube
from . import localfile as provider_localfile


# -*- coding: utf-8 -*-
'''basic submission tree model impl'''
from re import sub


class SubmissionVideos(list):
    '''Parent submission'''
    def extend(self, __iterable) -> None:
        for item in __iterable:
            self.append(item)

    def append(self, video) -> None:
        '''Only Submission will be appended to our list'''
        if isinstance(video,dict):
            # try to interpert it as a list of dictionaries sent by server            
            with Submission() as submission:                
                submission.video_endpoint = video['filename']
                submission.video_duration = video['duration']
                submission.title = video['title']            
                submission.bvid = video['bvid']
                submission.biz_id = video['cid']
                submission.aid = video['aid']
                submission.stat = video
                submission.parent = self
            return super().append(submission)            
        elif isinstance(video,Submission):
            return super().append(video)
        else:
            raise Exception("Either a dict or a Submission object can be supplied.")
            
    @property
    def archive(self):
        '''Dumps current videos as archvies that's to be the payload'''
        target = self if self else [self.parent] # fallback to parent node
        return [{
            "filename": video.video_endpoint,
            "title": video.title,
            "desc": video.description,
            "cid":video.biz_id
        } for video in target]

    def __init__(self,parent=None):
        '''Initializes the list
        
            parent : Submission - Used as fallback value when theres no subvideos
        '''
        self.parent = parent
        super().__init__()

class Submission:
    '''COPYRIGHT types'''
    COPYRIGHT_ORIGINAL = 1
    COPYRIGHT_REUPLOAD = 2
    '''REPRINT types'''
    REPRINT_DISALLOWED = 1
    REPRINT_ALLOWED = 0
    '''Copyright consts'''
    close_reply: bool = False
    close_danmu: bool = False
    '''Access control parameters'''    
    _description: str = ''
    @property
    def description(self):
        return self._description
    @description.setter
    def description(self,v):
        self._description = v or '' # fallback
    '''Description for the video'''
    title: str = ''
    '''Title for the submission'''
    copyright : int = COPYRIGHT_REUPLOAD
    '''Copyright type'''
    no_reprint : int = REPRINT_ALLOWED
    '''Reupload allowance type''' 
    source: str = ''
    '''Reupload source'''    
    thread: int = 19
    '''Thread ID'''
    tags: list = None
    '''Tags of video'''
    videos: SubmissionVideos = None
    '''List of videos in submission'''
    _cover_url = ''        
    @property
    def cover_url(self):
        '''Cover image URL'''
        return self._cover_url
    @cover_url.setter
    def cover_url(self,value):
        # note : this will strip the HTTP prefix        
        if value:self._cover_url = '//' + value.split('//')[-1]         
    # region Per video attributes    
    _video_filename = ''
    @property
    def video_endpoint(self):
        '''Endpoint name'''
        return self._video_filename
    @video_endpoint.setter
    def video_endpoint(self,value):
        # note : this will strip the HTTP prefix        
        self._video_filename = value.split('/')[-1].split('.')[0]    
    biz_id = 0
    '''a.k.a cid.for web apis'''        
    bvid = ''
    '''the new video ID'''
    aid = 0
    '''another ID for web apis'''
    thread_name = ''
    '''upload thread name i.e. typename'''
    parent_tname = ''
    '''parent thread name'''
    stat = None
    '''viewer status'''
    reject_reason = ''
    '''rejection'''
    state = 0
    '''status of video'''
    state_desc = ''
    '''status but human readable'''
    video_duration = 0
    '''duration of video'''
    _parent = None
    @property
    def parent(self):
        self._parent : Submission
        return self._parent
    @parent.setter
    def parent(self,v):
        self._parent = v
    '''parent object. used for videos property'''
    # endregion
    def __init__(self) -> None:
        self.tags = [] # creates new instance for mutables
        self.videos = SubmissionVideos(self)
    def __enter__(self):
        '''Creates a new,empty submission'''
        return Submission()
    def __exit__(self,*args):pass
    @property
    def archive(self):
        '''returns a dict containing all our info'''
        kv_pair = {
            "copyright": self.copyright,
            "videos": self.videos.archive,
            "source": self.source,
            "tid": int(self.thread),            
            "title": self.title,
            "tag": ','.join(set(self.tags)),
            # "desc_format_id": 31,
            "desc": self.description,
            # "up_close_reply": self.close_reply,
            # "up_close_danmu": self.close_danmu
            "no_reprint":self.no_reprint
        }        
        if self.cover_url:
            kv_pair = {**kv_pair,"cover": self.cover_url}
        return kv_pair
    def __repr__(self) -> str:
        return '< bvid : "%s" , title : "%s", desc : "%s" , video_endpoint : "%s" >' % (
            self.bvid,self.title,self.description,
            self.video_endpoint or ','.join([video.video_endpoint for video in self.videos])
        )
def create_submission_by_arc(arc : dict):
    '''Generates a `Submission` object via a `arc` dict'''
    with Submission() as submission:
        submission.stat = arc['stat'] if 'stat' in arc else arc['archive']
        submission.aid = submission.stat['aid']        
        if 'parent_tname' in arc: # Web version only
            submission.parent_tname = arc['parent_tname']
            submission.thread_name = arc['typename']
        if 'Archive' in arc:
            arc['archive'] = arc['Archive']                
        submission.copyright = arc['archive']['copyright']
        submission.bvid = arc['archive']['bvid']
        submission.title = arc['archive']['title']
        submission.cover_url = arc['archive']['cover']
        submission.tags = arc['archive']['tag'].split(',')
        submission.description = arc['archive']['desc']
        submission.source = arc['archive']['source']
        submission.state_desc = arc['archive']['state_desc']
        submission.state = arc['archive']['state']
        submission.reject_reason = arc['archive']['reject_reason']  
        submission.no_reprint = arc['archive']['no_reprint']
        if 'Videos' in arc:
            arc['videos'] = arc['Videos']
        submission.videos.extend(arc['videos'])
    return submission
